/*
	File:		MoreFinderEvents.h

	Contains:	Functions to help you build and sending Apple events to the Finder.

	Written by:	George Warner

	Copyright:	Copyright (c) 2000 Apple Computer, Inc., All Rights Reserved.

				You may incorporate this Apple sample source code into your program(s) without
				restriction. This Apple sample source code has been provided "AS IS" and the
				responsibility for its operation is yours. You are not permitted to redistribute
				this Apple sample source code as "Apple sample source code" after having made
				changes. If you're going to re-distribute the source, we require that you make
				it clear in the source that the code was descended from Apple sample source
				code, but that you've made changes.

	Change History (most recent first):

         <4>     4/26/00    gaw     Added MoreFEGetKind
         <3>      3/9/00    gaw     Fix AEDesc.dataHandle references
         <2>      3/9/00    gaw     API changes for MoreAppleEvents
*/

//*******************************************************************************
//	A private conditionals file to setup the build environment for this project.
#include "MoreSetup.h"
//**********	Universal Headers		****************************************
#include <Aliases.h>
#include <AEDataModel.h>
#include <ASRegistry.h>
#include <FinderRegistry.h>
#include <Folders.h>
#include <Icons.h>
#include <OSA.h>
#include <TextUtils.h>
//**********	Project Headers			****************************************
#include "MoreAppleEvents.h"
#include "MoreAEObjects.h"
#include "MoreFinderEvents.h"
#include "MoreProcesses.h"
/********************************************************************************
	Send an Apple event to the Finder to get file kind of the item 
	specified by the FSSpecPtr.

	pFSSpecPtr		==>		The item to get the file kind of.
	pKindStr		==>		A string into which the file kind will be returned.
	pKindStr		<==		the file kind of the FSSpec.
	pIdleProcUPP	==>		A UPP for an idle function, or nil.
	
	See note about idle functions above.
*/
pascal OSErr MoreFEGetKind(const FSSpecPtr pFSSpecPtr,Str255 pKindStr,const AEIdleUPP pIdleProcUPP)
{
	AppleEvent theEvent = {typeNull, nil};	//	If you always init AEDescs, it's always save to dispose of them.
	OSErr anErr = noErr;

	if (nil == pIdleProcUPP)	// the idle proc is required
		return paramErr;

	anErr = MoreAECreateAppleEventSignatureTarget(kFinderFileType, kFinderCreatorType,
											  kAECoreSuite, kAEGetData, &theEvent);
	if (anErr == noErr)
	{
		AEDesc 		itemObject = {typeNull, nil};
		AEDesc		containerObj = {typeNull, nil};		// start with the null (application) container
		
		anErr = MoreAEOCreateAliasObjectFromFSSpec(pFSSpecPtr, &containerObj, &itemObject);
		if (anErr == noErr)
		{
			AEDesc 		propertyObject = {typeNull, nil};
			
			anErr = MoreAEOCreatePropertyObject(pKind, &itemObject, &propertyObject);
			AEDisposeDesc(&itemObject);
			if (anErr == noErr)
			{
				anErr = AEPutParamDesc(&theEvent, keyDirectObject, &propertyObject);
				AEDisposeDesc(&propertyObject);
				if (anErr == noErr)
				{
					anErr = MoreAESendEventReturnPString(pIdleProcUPP,&theEvent,pKindStr);
				}
			}
		}
		(void) AEDisposeDesc(&theEvent);	// always dispose of AEDescs when you are finished with them
	}
	return anErr;
}
/********************************************************************************
	Sets the Finder's selection to nothing by telling it to set it's 
	selection to an empty list.

	pIdleProcUPP 	==>		nil for default, or UPP for the idle function to use
	
	Requires that the folder's window be open, otherwise an error is returned
	in the reply event.
	
	See note about idle functions above.
*/
pascal OSErr MoreFESetSelectionToNone(const AEIdleUPP pIdleProcUPP)
{
	OSErr		anErr = noErr;
	
	AppleEvent	theEvent = {typeNull, nil};	//	If you always init AEDescs, it's always save to dispose of them.
	
	anErr = MoreAECreateAppleEventSignatureTarget(kFinderFileType, kFinderCreatorType,
											  kAECoreSuite, kAESetData, &theEvent);
	if (anErr == noErr)
	{
		AEDesc		containerObj = {typeNull, nil};	// start with the null (application) container
		AEDesc 		propertyObject = {typeNull, nil};
			
		anErr = MoreAEOCreatePropertyObject(pSelection, &containerObj, &propertyObject);
		if (anErr == noErr)
		{
			anErr = AEPutParamDesc(&theEvent, keyDirectObject, &propertyObject);
			AEDisposeDesc(&propertyObject);	//	Always dispose of objects as soon as you are done (helps avoid leaks)
			if (anErr == noErr)
			{
				AEDescList		emptyList = {typeNull, nil};
				
				anErr = AECreateList(nil, 0, false, &emptyList);
				if (anErr == noErr)
				{
					anErr = AEPutParamDesc(&theEvent, keyAEData, &emptyList);
					if (anErr == noErr)
					{
						anErr = MoreAESendEventNoReturnValue(pIdleProcUPP, &theEvent);
					}
				}
			}
		}
		(void) AEDisposeDesc(&theEvent);	// always dispose of AEDescs when you are finished with them
	}
	return anErr;
} // MoreFESetSelectionToNone
/********************************************************************************
	Send an Apple event to Finder 7.5 or later to set the view of the window
	for the folder pointed to by the FSSpec.

	pFSSpecPtr		==>		FSSpec for the folder whose view is to be set
	pViewStyle		==>		A view constant, as defined in AERegistry.h
	pIdleProcUPP 	==>		nil for default, or UPP for the idle function to use
	
	Requires that the folder's window be open, otherwise an error is returned
	in the reply event.
	
	See note about idle functions above.
*/
pascal OSErr MoreFEChangeFolderViewNewSuite(const FSSpecPtr pFSSpecPtr,
										  const long pViewStyle,
										  const AEIdleUPP pIdleProcUPP)
{
	OSErr		anErr = noErr;
	
	AppleEvent	theEvent = {typeNull, nil};
	
	anErr = MoreAECreateAppleEventSignatureTarget(kFinderFileType, kFinderCreatorType,
											  kAECoreSuite, kAESetData, &theEvent);
	if (anErr == noErr)
	{
		AEDesc 		folderObject = {typeNull, nil};
		AEDesc		containerObj = {typeNull, nil};	// start with the null (application) container
		
		anErr = MoreAEOCreateAliasObjectFromFSSpec(pFSSpecPtr, &containerObj, &folderObject);
		if (anErr == noErr)
		{
			AEDesc 		propertyObject = {typeNull, nil};
			
			anErr = MoreAEOCreatePropertyObject(pView, &folderObject, &propertyObject);
			AEDisposeDesc(&folderObject);
			if (anErr == noErr)
			{
				anErr = AEPutParamDesc(&theEvent, keyDirectObject, &propertyObject);
				AEDisposeDesc(&propertyObject);
				if (anErr == noErr)
				{
					anErr = AEPutParamPtr(&theEvent, keyAEData,
										   typeLongInteger, &pViewStyle, sizeof(pViewStyle));
					if (anErr == noErr)
					{
						anErr = MoreAESendEventNoReturnValue(pIdleProcUPP, &theEvent);
					}
				}
			}
		}
		(void) AEDisposeDesc(&theEvent);	// always dispose of AEDescs when you are finished with them
	}
	return anErr;
} // MoreFEChangeFolderViewNewSuite
/********************************************************************************
	Send an Apple event to Finder 7.1.1 or 7.1.2 to set the view of the window
	for the folder pointed to by the FSSpec.
	
	pFSSpecPtr		==>		FSSpec for the folder whose view is to be set
	pViewStyle		==>		A view constant, as defined in AERegistry.h
	pIdleProcUPP 	==>		nil for default, or UPP for the idle function to use
	
	Requires that the folder's window be open, otherwise an error is returned
	in the reply event.
	
	See note about idle functions above.
*/
pascal OSErr MoreFEChangeFolderViewOldSuite(const FSSpecPtr pFSSpecPtr,
										  const long pViewStyle,
										  const AEIdleUPP pIdleProcUPP)
{
	OSErr		anErr = noErr;
	
	AppleEvent	theEvent = {typeNull, nil};
	
	anErr = MoreAECreateAppleEventSignatureTarget(kFinderFileType, kFinderCreatorType,
											  kAEFinderEvents, kAEChangeView, &theEvent);
	if (anErr == noErr)
	{
		anErr = MoreAEOAddAliasParameterFromFSSpec(pFSSpecPtr, keyDirectObject, &theEvent);
		if (anErr == noErr)
		{
			anErr = AEPutParamPtr(&theEvent, keyMiscellaneous,
								   typeLongInteger, &pViewStyle, sizeof(pViewStyle));
			if (anErr == noErr)
			{
				anErr = MoreAESendEventNoReturnValue(pIdleProcUPP, &theEvent);
			}
		}
		(void) AEDisposeDesc(&theEvent);	// always dispose of AEDescs when you are finished with them
	}
	return anErr;
} // MoreFEChangeFolderViewOldSuite
/********************************************************************************
	Send an Apple event to the Finder (any version) to set the view of the window
	for the folder pointed to by the FSSpec.
	
	pFSSpecPtr		==>		FSSpec for the folder whose view is to be set
	pViewStyle		==>		A view constant, as defined in AERegistry.h
	pIdleProcUPP 	==>		nil for default, or UPP for the idle function to use
	
	Requires that the folder's window be open, otherwise an error is returned
	in the reply event.
	
	See note about idle functions above.
*/
pascal OSErr MoreFEChangeFolderView(const FSSpecPtr pFSSpecPtr,
								  const long pViewStyle,
								  const AEIdleUPP pIdleProcUPP)
{
	OSErr	anErr = noErr;
	
	if (MoreFECallsAEProcess())
	{
		if (MoreFEIsOSLCompliant())
		{
			anErr = MoreFEChangeFolderViewNewSuite(pFSSpecPtr, pViewStyle, pIdleProcUPP);
		}
		else
		{
			anErr = MoreFEChangeFolderViewOldSuite(pFSSpecPtr, pViewStyle, pIdleProcUPP);
		}
	}
	else
	{
		anErr = errAEEventNotHandled;
	}
	
	return anErr;
}//end MoreFEChangeFolderView
/********************************************************************************
	Send an Apple event to the Finder to add a custom icon to the item 
	specified by the fssPtr.
	
	pFSSpecPtr		==>		The item to add the custom icon to.
	pIconSuiteHdl	==>		A handle to the icon suite to install.
	pIconSelector	==>		An IconSelectorValue specifying which icon types
							to add (defined in Icons.h).
	pIdleProcUPP	==>		A UPP for an idle function, or nil.
	
	See note about idle functions above.
*/
pascal OSErr MoreFEAddCustomIconToItem(const FSSpecPtr pFSSpecPtr,
									 const Handle pIconSuiteHdl,
									 const IconSelectorValue pIconSelector,
									 const AEIdleUPP pIdleProcUPP)
{
	OSErr		anErr = noErr;
	
	AppleEvent	theEvent = {typeNull, nil};
	
	anErr = MoreAECreateAppleEventSignatureTarget(kFinderFileType, kFinderCreatorType,
											  kAECoreSuite, kAESetData, &theEvent);
	if (anErr == noErr)
	{
		AEDesc 		itemObject = {typeNull, nil};
		AEDesc		containerObj = {typeNull, nil};		// start with the null (application) container
		
		anErr = MoreAEOCreateAliasObjectFromFSSpec(pFSSpecPtr, &containerObj, &itemObject);
		if (anErr == noErr)
		{
			AEDesc 		propertyObject = {typeNull, nil};
			
			anErr = MoreAEOCreatePropertyObject(pIconBitmap, &itemObject, &propertyObject);
			AEDisposeDesc(&itemObject);
			if (anErr == noErr)
			{
				anErr = AEPutParamDesc(&theEvent, keyDirectObject, &propertyObject);
				AEDisposeDesc(&propertyObject);
				if (anErr == noErr)
				{
					AEDescList	iconFamilyRec = {typeNull, nil};
					
					anErr = MoreFECreateIconFamilyRecord(pIconSuiteHdl, pIconSelector, &iconFamilyRec);
					if (anErr == noErr)
					{
						anErr = AEPutParamDesc(&theEvent, keyAEData, &iconFamilyRec);
						if (anErr == noErr)
						{
							anErr = MoreAESendEventNoReturnValue(pIdleProcUPP, &theEvent);
						}
					}
				}
			}
		}
		(void) AEDisposeDesc(&theEvent);	// always dispose of AEDescs when you are finished with them
	}
	return anErr;
} // MoreFEAddCustomIconToItem
/********************************************************************************
	Send an Apple event to the Finder to get the icon of the item 
	specified by the fssPtr.

	pFSSpecPtr		==>		The item to get the icon from.
	pIdleProcUPP	==>		A UPP for an idle function, or nil.
	pIconSuiteHdl	==>		A handle into which the icon suite will be returned.
	pIconSuiteHdl	<==		A handle to an icon suite.
	
	See note about idle functions above.
*/
pascal OSErr MoreFEGetItemIconSuite(const FSSpecPtr pFSSpecPtr,
								  const AEIdleUPP pIdleProcUPP,
										Handle	 *pIconSuiteHdl)
{
	OSErr	anErr = noErr;
	
	AppleEvent	theEvent = {typeNull, nil};
	
	anErr = MoreAECreateAppleEventSignatureTarget(kFinderFileType, kFinderCreatorType,
											  kAECoreSuite, kAEGetData, &theEvent);
	if (anErr == noErr)
	{
		AEDesc 		itemObject = {typeNull, nil};
		AEDesc		containerObj = {typeNull, nil};		// start with the null (application) container
		
		anErr = MoreAEOCreateAliasObjectFromFSSpec(pFSSpecPtr, &containerObj, &itemObject);
		if (anErr == noErr)
		{
			AEDesc 		propertyObject = {typeNull, nil};
			
			anErr = MoreAEOCreatePropertyObject(pIconBitmap, &itemObject, &propertyObject);
			AEDisposeDesc(&itemObject);
			if (anErr == noErr)
			{
				anErr = AEPutParamDesc(&theEvent, keyDirectObject, &propertyObject);
				AEDisposeDesc(&propertyObject);
				if (anErr == noErr)
				{
					AppleEvent	theReply = {typeNull, nil};
					AESendMode	sendMode;
					
					if (pIdleProcUPP == nil)
						sendMode = kAENoReply;
					else
						sendMode = kAEWaitReply;

					anErr = AESend(&theEvent, &theReply, sendMode, kAENormalPriority,
									kNoTimeOut, pIdleProcUPP, nil);
					if (anErr == noErr)
					{
						anErr =  MoreAEGetHandlerError(&theReply);
						if (anErr == noErr)
						{
							AEDesc	iconFamilyRec = {typeNull, nil};
							
							anErr = AEGetParamDesc(&theReply, keyDirectObject, typeWildCard, &iconFamilyRec);
							if (anErr == noErr)
							{
								anErr = MoreFECreateIconSuite(&iconFamilyRec, pIconSuiteHdl);
							}
							AEDisposeDesc(&iconFamilyRec);
						}
						AEDisposeDesc(&theReply);
					}
				}
			}
		}
		(void) AEDisposeDesc(&theEvent);	// always dispose of AEDescs when you are finished with them
	}
	return anErr;
} // MoreFEGetItemIconSuite
/********************************************************************************
	Send an Apple event to the Finder to get a list of Finder-style object
	for each item on the desktop. This includes files (of all types),
	folders, and volumes.
	
	pIdleProcUPP	==>		A UniversalProcPtr for an idle function, or nil.
	pObjectList		==>		A null AEDesc.
					<==		A list containing object descriptors,
							or a null AEDesc if an error is encountered.
	
	See note about idle functions above.
*/
pascal OSErr MoreFEGetEveryItemOnDesktop(const AEIdleUPP pIdleProcUPP,
											 AEDescList *pObjectList)
{
	OSErr		anErr = noErr;
	
	AppleEvent	theEvent = {typeNull, nil};
	
	pObjectList->descriptorType = typeNull;
	pObjectList->dataHandle = nil;
	
	anErr = MoreAECreateAppleEventSignatureTarget(kFinderFileType, kFinderCreatorType,
											  kAECoreSuite, kAEGetData, &theEvent);
	if (anErr == noErr)
	{
		AEDesc		containerObj = {typeNull, nil};		// start with the null (application) container
		AEDesc 		propertyObject = {typeNull, nil};
		
		anErr = MoreAEOCreatePropertyObject(kDesktopFolderType, &containerObj, &propertyObject);
		if (anErr == noErr)
		{
			DescType	selectAll = kAEAll;
			AEDesc		allObjectsDesc = {typeNull, nil};
			
			anErr = MoreAEOCreateSelectionObject(selectAll, &propertyObject, &allObjectsDesc);
			AEDisposeDesc(&propertyObject);
			if (anErr == noErr)
			{
				anErr = AEPutParamDesc(&theEvent, keyDirectObject, &allObjectsDesc);
				AEDisposeDesc(&allObjectsDesc);
				if (anErr == noErr)
				{
					AppleEvent	theReply = {typeNull, nil};
					AESendMode	sendMode;
					
					if (pIdleProcUPP == nil)
						sendMode = kAENoReply;
					else
						sendMode = kAEWaitReply;

					anErr = AESend(&theEvent, &theReply, sendMode, kAENormalPriority,
									kNoTimeOut, pIdleProcUPP, nil);
					if (anErr == noErr)
					{
						anErr =  MoreAEGetHandlerError(&theReply);
						
						if (!anErr && theReply.descriptorType != typeNull)
						{
							anErr = AEGetParamDesc(&theReply, keyDirectObject, typeAEList, pObjectList);
						}
						AEDisposeDesc(&theReply);
					}
				}
			}
		}
		(void) AEDisposeDesc(&theEvent);	// always dispose of AEDescs when you are finished with them
	}
	return anErr;
} // MoreFEGetEveryItemOnDesktop
/********************************************************************************
	Send an Apple event to the Finder to get the current font and font size
	as set in the Views control panel.
	
	pIdleProcUPP	==>		A UniversalProcPtr for an idle function, or nil.
	pFont			==>		The font ID for the current view settings.
	pFontSize		==>		The font size for the current view settings.
	
	You MUST supply an idle function for this routine to work, since it is
	returning data.  You may use the simple handler provided in this library.
	
	See note about idle functions above.
	
	Errors
	
	-50		paramErr	No idle function provided.
*/
pascal OSErr MoreFEGetViewFontAndSize(const AEIdleUPP pIdleProcUPP,
										  SInt16 *pFont,
										  SInt16 *pFontSize)
{
	OSErr		anErr = noErr;
	
	AppleEvent	theEvent = {typeNull, nil};
		
	if (pIdleProcUPP == nil)
	{
		return paramErr;
	}
	//	Build the event we will send
	anErr = MoreAECreateAppleEventSignatureTarget(kFinderFileType, kFinderCreatorType,
											  kAECoreSuite, kAEGetData, &theEvent);
	if (anErr == noErr)
	{
		AEDesc		containerObj = {typeNull, nil};		// start with the null (application) container
		AEDesc 		viewPropertyObj = {typeNull, nil};
		//	Create an object descriptor for the view preferences property
		anErr = MoreAEOCreatePropertyObject(pViewPreferences, &containerObj, &viewPropertyObj);
		if (anErr == noErr)
		{
			AEDesc 		fontPropertyObj = {typeNull, nil};
			
			//	Create an object descriptor for the view font property of the view preferences property
			anErr = MoreAEOCreatePropertyObject(pViewFont, &viewPropertyObj, &fontPropertyObj);
			//	Don't dispose of the fontPropertyObj yet, we will need it for the font size property
			if (anErr == noErr)
			{
				anErr = AEPutParamDesc(&theEvent, keyDirectObject, &fontPropertyObj);
				(void) AEDisposeDesc(&fontPropertyObj);
				if (anErr == noErr)
				{
					anErr = MoreAESendEventReturnSInt16(pIdleProcUPP, &theEvent, pFont);
					if (anErr == noErr)
					{
						AEDesc 		fontSizePropertyObj = {typeNull, nil};
						
						anErr = MoreAEOCreatePropertyObject(pViewFontSize, &viewPropertyObj, &fontSizePropertyObj);
						(void) AEDisposeDesc(&viewPropertyObj);	// always dispose of AEDescs when you are finished with them
						if (anErr == noErr)
						{
							anErr = AEPutParamDesc(&theEvent, keyDirectObject, &fontSizePropertyObj);
							(void) AEDisposeDesc(&fontSizePropertyObj);
							if (anErr == noErr)
							{
								anErr = MoreAESendEventReturnSInt16(pIdleProcUPP, &theEvent, pFontSize);
							}
						}
					}
				}
			}
		}
		(void) AEDisposeDesc(&theEvent);	// always dispose of AEDescs when you are finished with them
	}
	return anErr;
} // MoreFEGetEveryItemOnDesktop
/********************************************************************************
	Send an Apple event to the Finder to update the display of the item specified
	by the FSSpec.

	pFSSpecPtr		==>		The item whose display should be updated.

	No reply is returns, so none is asked for. Hence, no idle function is needed.
*/
pascal OSErr MoreFEUpdateItemFSS(const FSSpecPtr pFSSpecPtr)
{
	OSErr			anErr = noErr;
	AliasHandle		aliasHandle;
	
	anErr = NewAlias(nil, pFSSpecPtr, &aliasHandle);

	if (anErr == noErr)
	{
		anErr = MoreFEUpdateItemAlias(aliasHandle);
	}
	
	DisposeHandle((Handle)aliasHandle);
	
	return anErr;
} // MoreFEChangeFolderView
/********************************************************************************
	Send an Apple event to the Finder (any version) to set the visibility of
	a process specified by it's PSN.
	
	pPSN			==>		The processes serial number
	pVisible		==>		Make the process visible or not
	
*/
pascal OSErr MoreFESetProcessVisibility(const ProcessSerialNumberPtr pPSN, Boolean pVisible)
{
	OSErr		anErr = noErr;
	AppleEvent	theEvent = {typeNull, nil};
	
	anErr = MoreAECreateAppleEventSignatureTarget(kFinderFileType, kFinderCreatorType,
											  kAECoreSuite, kAESetData, &theEvent);
	if (anErr == noErr)
	{
		OSType		processType;
		OSType		processSignature;
		
		anErr = MoreProcGetProcessTypeSignature(pPSN, &processType, &processSignature);
		if (anErr == noErr)
		{
			AEDesc 	processObject = {typeNull, nil};
			
		//	To hide the Finder it's visible property is directly set to false -- no process target
		//	is needed.  Only create a non-null processObject if the Finder is not the target.
			if ((processSignature != kFinderProcessSignature)  ||  (processType != kFinderProcessType))
			{	
				Str32	processName;
				
			//	The Finder expects the process to be specified by name.
				anErr = MoreProcGetProcessName(pPSN, processName);
				if (anErr == noErr)
				{
					AEDesc	containerObj = {typeNull, nil};
					AEDesc	nameDesc = {typeNull, nil};
					
					anErr = AECreateDesc(typeChar, &processName[ 1 ], processName[ 0 ], &nameDesc);
					if (anErr == noErr)
					{
						anErr = CreateObjSpecifier(cProcess, &containerObj, formName,
													&nameDesc, false, &processObject);
						(void) AEDisposeDesc(&nameDesc);
					}
				}
			}
			
			if (anErr == noErr)
			{
				AEDesc	propertyObject = {typeNull, nil};

			//	To hide a process tell the Finder to set the process' visible property to false.
				anErr = MoreAEOCreatePropertyObject(pVisible, &processObject, &propertyObject);
				(void) AEDisposeDesc(&processObject);
				if (anErr == noErr)
				{
					anErr = AEPutParamDesc(&theEvent, keyDirectObject, &propertyObject);
					(void) AEDisposeDesc(&propertyObject);
					if (anErr == noErr)
					{
						anErr = AEPutParamPtr(&theEvent, keyAEData, typeBoolean, &pVisible, sizeof(Boolean));
						if (anErr == noErr)
						{
							anErr = MoreAESendEventNoReturnValue(nil, &theEvent);
						}
					}
				}
			}
		}
		(void) AEDisposeDesc(&theEvent);	// always dispose of AEDescs when you are finished with them
	}
	return anErr;
} // MoreFEChangeFolderView
/********************************************************************************
	Send an Apple event to the Finder to update the display of the item specified
	by the alias.

	pAliasHdl		==>		The item whose display should be updated.

	No reply is returns, so none is asked for. Hence, no idle function is needed.
*/
pascal OSErr MoreFEUpdateItemAlias(const AliasHandle pAliasHdl)
{
	OSErr		anErr = noErr;
	
	AppleEvent	theEvent = {typeNull, nil};
	
	anErr = MoreAECreateAppleEventSignatureTarget(kFinderFileType, kFinderCreatorType,
											  kAEFinderSuite, kAESync, &theEvent);
	if (anErr == noErr)
	{
		AEDesc 		aliasDesc = {typeNull, nil};
		
		anErr = MoreAEOCreateAliasDesc(pAliasHdl, &aliasDesc);
		if (anErr == noErr)
		{
			anErr = AEPutParamDesc(&theEvent, keyDirectObject, &aliasDesc);
			AEDisposeDesc(&aliasDesc);
			if (anErr == noErr)
			{
				anErr = MoreAESendEventNoReturnValue(nil, &theEvent);
			}
		}
		(void) AEDisposeDesc(&theEvent);	// always dispose of AEDescs when you are finished with them
	}
	return anErr;
} // MoreFEUpdateItemAlias
/********************************************************************************
	Send an odoc Apple event to the Finder to open the item specified by the FSSpec.
	
	This routine can be used to open a file with it's creator app, launch an,
	open a control panel. Pretty much open anything you can open directly in the
	Finder by double-clicking.
	
	pFSSpecPtr		==>		The item whose display should be updated.

	No reply is returns, so none is asked for. Hence, no idle function is needed.
*/
pascal	OSErr MoreFEOpenFile(const FSSpec *pFSSpecPtr)
{
	OSErr		anErr = noErr;
	
	AppleEvent	theEvent = {typeNull, nil};
	
	anErr = MoreAECreateAppleEventSignatureTarget(kFinderFileType, kFinderCreatorType,
											  kAECoreSuite, kAEOpen, &theEvent);
	if (anErr == noErr)
	{
		anErr = AEPutParamPtr(&theEvent, keyDirectObject, typeFSS, pFSSpecPtr, sizeof(FSSpec));
		if (anErr == noErr)
		{	
			anErr = MoreAESendEventNoReturnValue(nil, &theEvent);
		}
		(void) AEDisposeDesc(&theEvent);	// always dispose of AEDescs when you are finished with them
	}
	return anErr;
}//end MoreFEOpenFile
/********************************************************************************
	Send an Apple event to the Finder to create a new alias file at destFSSPtr
	location, with the file at sourceFSSPtr as the target, with newName as it's name.
	
	pSourceFSSPtr	==>		The target for the new alias file.
	pDestFSSPtr		==>		The location for the new alias file.
	pNewName			==>		The name for the new alias file.
	pIdleProcUPP	==>		A UniversalProcPtr for an idle function, or nil.

	A reference to the newly created alias file will be returned. Currently this
	function ignores this result, but you could extract it if needed. You can ask
	that the result be returned as a Finder style object, as an alias, or as an FSSpec.

	See note about idle functions above.
*/
pascal	OSErr	MoreFECreateAliasFile(const FSSpecPtr pSourceFSSPtr,
								  const FSSpecPtr pDestFSSPtr,
								  ConstStr63Param pNewName,
								  const AEIdleUPP pIdleProcUPP)
{
	OSErr		anErr = noErr;
	
	AppleEvent	theEvent = {typeNull, nil};
	
	anErr = MoreAECreateAppleEventSignatureTarget(kFinderFileType, kFinderCreatorType,
											  kAECoreSuite, kAECreateElement, &theEvent);
	if (anErr == noErr)
	{
		//	put the type of item to make into the Apple event
		OSType	anAliasType = typeAlias;
		
		anErr = AEPutParamPtr(&theEvent, keyAEObjectClass, typeType, &anAliasType, sizeof(OSType));
		if (anErr == noErr)
		{
			//	add an alias that points to the targer for the alias file to be created
			anErr = MoreAEOAddAliasParameterFromFSSpec(pSourceFSSPtr, keyASPrepositionTo, &theEvent);
			if (anErr == noErr)
			{
				//	where should the alias file be created?
				anErr = MoreAEOAddAliasParameterFromFSSpec(pDestFSSPtr, keyAEInsertHere, &theEvent);
				if (anErr == noErr)
				{
					//	add a properties record so we can specify the name of the new alias file.
					AERecord	properties = {typeNull, nil};
					
					anErr = AECreateList(nil, 0, true, &properties);
					if (anErr == noErr)
					{
						anErr = AEPutKeyPtr(&properties, keyAEName, typeChar, pNewName+1, *pNewName);
						if (anErr == noErr)
						{
							anErr = AEPutParamDesc(&theEvent, keyAEPropData, &properties);
							if (anErr == noErr)
							{
								//	Send the event. In this case we don't care about the reply, but if we did we
								//	would get back an object that describes the alias file just created
								//	(either an object descriptor, an alias, or an FSSpec, depending on what we ask for).
								anErr = MoreAESendEventNoReturnValue(pIdleProcUPP, &theEvent);
							}
						}
						AEDisposeDesc(&properties);
					}
				}
			}
		}
		(void) AEDisposeDesc(&theEvent);	// always dispose of AEDescs when you are finished with them
	}
	return anErr;
}//end MoreFECreateAliasFile
/********************************************************************************
	Send set data Apple event to the Finder to change the position of the icon
	for the item specified by the FSSpec.  The Finder's display of the item is
	immediately updated.
	
	NOTE:	The position parameter is a Point, with it's values in y/x order,
			where as the Finder expects positions to be in x/y order.  This
			routine changes the order for you. 
	
	pFSSpecPtr		==>		The item whose position will be changed.
	pPosition		==>		The new position for the item.

	No reply is returns, so none is asked for. Hence, no idle function is needed.
*/
pascal OSErr MoreFEMoveDiskIcon(const FSSpecPtr pFSSpecPtr,
							  const Point pPoint)
{
	OSErr		anErr = noErr;
	
	AppleEvent	theEvent = {typeNull, nil};
	
	anErr = MoreAECreateAppleEventSignatureTarget(kFinderFileType, kFinderCreatorType,
											  kAECoreSuite, kAESetData, &theEvent);
	if (anErr == noErr)
	{
		AEDesc 		itemObject = {typeNull, nil};
		AEDesc		containerObj = {typeNull, nil};	// start with the null (application) container
		
		anErr = MoreAEOCreateAliasObjectFromFSSpec(pFSSpecPtr, &containerObj, &itemObject);
		if (anErr == noErr)
		{
			AEDesc 		propertyObject = {typeNull, nil};
			
			anErr = MoreAEOCreatePropertyObject(pPosition, &itemObject, &propertyObject);
			AEDisposeDesc(&itemObject);
			if (anErr == noErr)
			{
				anErr = AEPutParamDesc(&theEvent, keyDirectObject, &propertyObject);
				AEDisposeDesc(&propertyObject);
				if (anErr == noErr)
				{
					AEDescList	posList;
					
					anErr = MoreAEOCreatePositionList(pPoint, &posList);
					if (anErr == noErr)
					{
						anErr = AEPutParamDesc(&theEvent, keyAEData, &posList);
						if (anErr == noErr)
						{	
							anErr = MoreAESendEventNoReturnValue(nil, &theEvent);
						}
					}
				}
			}
		}
		(void) AEDisposeDesc(&theEvent);	// always dispose of AEDescs when you are finished with them
	}
	return anErr;
} // MoreFEChangeFolderView
/********************************************************************************
	Make an icon suite containing the icons in an icon family record, as returned
	by the Finder. Behaves simmilar to a call to GetIconSuite, i.e., a new icon
	suite handle will be returned in pIconSuiteHdl.
	
	pIconFamilyAEDescList	input:	The icon family to process.
	pIconSuiteHdl		input:	Pointer to an icon suite handle variable
						output:	An icon suite containing the icons from the
								icon family record.
	
	Result Codes
	____________
	noErr				    0	No error	
	paramErr			  -50	The value of target or alias parameter, or of
								both, is NIL, or the alias record is corrupt
	memFullErr			 -108	Not enough room in heap zone	
	errAECoercionFail 	-1700	Data could not be coerced to the requested 
								Apple event data type	
*/
pascal	OSErr	MoreFECreateIconSuite(const AEDescList *pIconFamilyAEDescList,
										Handle *pIconSuiteHdl)
{
	OSErr	anErr = noErr;

	const	long		iconTypesCnt = 6;
	static	DescType	iconTypes[] = {	typeIconAndMask,
										type8BitIcon,
										type4BitIcon,
										typeSmallIconAndMask,
										typeSmall8BitIcon,
										typeSmall4BitIcon};
	
	AEDescList	iconList = {typeNull, nil};
	
	if (MoreFEUsesIconFamily())
	{
		anErr = AECoerceDesc(pIconFamilyAEDescList, typeAERecord, &iconList);
	}
	else
	{
		iconList = *pIconFamilyAEDescList;
	}
	
	if (anErr == noErr)
	{
		anErr = NewIconSuite(pIconSuiteHdl);
		if (anErr == noErr)
		{
			long	index;
			AEDesc	iconDataDesc = {typeNull, nil};
			
			for (index = 0; index < iconTypesCnt; index++)
			{
				anErr = AEGetKeyDesc(&iconList, iconTypes[ index ],
									  typeWildCard, &iconDataDesc);
				if (anErr == noErr)
				{
					anErr = AddIconToSuite((void*) iconDataDesc.dataHandle,*pIconSuiteHdl,iconTypes[index]);
				}
			}
		}
	}
	AEDisposeDesc(&iconList);

	return anErr;
}//end MoreFECreateIconSuite
/********************************************************************************
	Used by MoreFECreateIconFamilyRecord, passed to ForEachIconDo as the IconAction
	function. Puts each icon in an icon suite into the descriptor record passed
	in the myDataPtr parameter.
*/
static pascal OSErr MyIconAction(ResType theIconType, Handle *theIcon, void *myDataPtr)
{
	OSErr	anErr = noErr;
	
	if (*theIcon != nil)	// only add the icon if it's really there
	{
		AEDescList *pIconFamilyAEDescList = (AEDescList*)myDataPtr;
		
		anErr = AEPutKeyPtr(pIconFamilyAEDescList, theIconType, theIconType,
							 **theIcon, GetHandleSize(*theIcon));
	}
		
	return anErr;
}//end MyIconAction
/********************************************************************************
	Make an icon family record containing the icons specified in the
	pIconSelector parameter.
	The pIconSuiteHdl parameter should contain an icon suite, as returned by a
	call to GetIconSuite.
	
	pIconSuiteHdl			input:	The icon suite to build the record from.
	pIconSelector			input:	Which icons to include in the record.
	pIconFamilyAEDescList	input:	Pointer to null AEDesc.
							output:	An AERecord that's been coerced to an
								icon family record.
	
	Result Codes
	____________
	noErr				    0	No error	
	paramErr			  -50	The value of target or alias parameter, or of
								both, is NIL, or the alias record is corrupt
	memFullErr			 -108	Not enough room in heap zone	
	errAECoercionFail 	-1700	Data could not be coerced to the requested 
								Apple event data type	
*/
pascal	OSErr	MoreFECreateIconFamilyRecord(	const Handle pIconSuiteHdl,
												const IconSelectorValue pIconSelector,
												AEDescList *pIconFamilyAEDescList)
{
	OSErr	anErr = noErr;
	AEDescList	iconList = {typeNull, nil};
	
	static	IconActionUPP iconActionUPP;
	
	if (iconActionUPP == nil)
		iconActionUPP = NewIconActionProc(&MyIconAction);

	// create a record for the icon family
	anErr = AECreateList(nil, 0, true, &iconList);
	
	if (anErr == noErr)
	{
		ForEachIconDo(pIconSuiteHdl, pIconSelector,iconActionUPP, &iconList);
		
		if (MoreFEUsesIconFamily())
		{
			anErr = AECoerceDesc(&iconList, typeIconFamily, pIconFamilyAEDescList);
		}
		else
		{
			*pIconFamilyAEDescList = iconList;
		}
		
		AEDisposeDesc(&iconList);
	}

	return anErr;
}//end MoreFECreateIconFamilyRecord
/********************************************************************************
	Does the Finder call AEProcessAppleEvent when it receives an Apple event.
	
	RESULT CODES
	____________
	true	Finder is version 7.1.3 or later, calls AEProcessAppleEvent,
			and supports the full old Finder event suite.
	false	The Finder supports a subset of the old Finder event suite.
	
	Use this routine together with MoreFEIsOSLCompliant to determine which
	suite of events the Finder supports.
*/
pascal Boolean MoreFECallsAEProcess(void)
{
	OSErr	anErr = noErr;
	
	static	long		gMFECallsAEProcess = kFlagNotSet;
		
	if (gMFECallsAEProcess == kFlagNotSet)
	{
		long	response;
		
		if (Gestalt(gestaltFinderAttr, &response) == noErr)
		{
			gMFECallsAEProcess = (response & (1L << gestaltFinderCallsAEProcess)) != 0;
		}
	}
	
	return gMFECallsAEProcess;
}//end MoreFECallsAEProcess
/********************************************************************************
	Does the Finder uses the ObjectSupportLib to resolve objecs.
		
	RESULT CODES
	____________
	true	Finder is version 7.5 or later, and supports the Standard event suite,
			the new Finder event suite & a subset of the old Finder event suite.
	false	See result of MoreFECallsAEProcess.
	
	Support for the old Finder event suite is limited, with some events missing
	and some events only partially supported. The subset of old Finder event
	supported is not the same subset as the original Finder 7 supported.
	
	Use the new event suite, and avoid the old one, whenever possible. 
*/
pascal Boolean MoreFEIsOSLCompliant(void)
{
	OSErr	anErr = noErr;
	
	static	long		gMFEIsOSLCompliant = kFlagNotSet;
	
	if (gMFEIsOSLCompliant == kFlagNotSet)
	{
		long	response;
		
		if (Gestalt(gestaltFinderAttr, &response) == noErr)
		{
			gMFEIsOSLCompliant = (response & (1L << gestaltOSLCompliantFinder)) != 0;
		}
	}
	return gMFEIsOSLCompliant;
}//end MoreFEIsOSLCompliant
/********************************************************************************
	Does the Finder uses cIconFamily records or IconSuites.
	
	So far (as of Mac OS 8.1) only the Finder in Mac OS 8.0 requires the use of
	IconSuites, rather than cIconFamily like the other Finders.  This test
	lets us identify this odd Finder so we can special case Icon code.
		
	RESULT CODES
	____________
	true	Finder uses cIconFamily.
	false	Finder uses IconSuite.
*/
pascal Boolean MoreFEUsesIconFamily(void)
{
	OSErr	anErr = noErr;
	
	static	long		gMFEUsesIconFamily = kFlagNotSet;
	
	if (gMFEUsesIconFamily == kFlagNotSet)
	{
		long	response;
		
		if (Gestalt(gestaltSystemVersion, &response) == noErr)
		{
			gMFEUsesIconFamily = (response != 0x00000800);
		}
	}
	
	return gMFEUsesIconFamily;
}//end MoreFEUsesIconFamily
//********************************************************************************
