/*
	File:		MoreOSA.c

	Contains:	Functions to help you when you are working with the OSA.

	Written by:	George Warner

	Copyright:	Copyright (c) 2000 Apple Computer, Inc., All Rights Reserved.

				You may incorporate this Apple sample source code into your program(s) without
				restriction. This Apple sample source code has been provided "AS IS" and the
				responsibility for its operation is yours. You are not permitted to redistribute
				this Apple sample source code as "Apple sample source code" after having made
				changes. If you're going to re-distribute the source, we require that you make
				it clear in the source that the code was descended from Apple sample source
				code, but that you've made changes.

	Change History (most recent first):

         <4>      3/9/00    gaw     Fixed OpaqueAEDataStorage.dataHandle references in
                                    MoreOSAUnloadScriptResource for CARBON
         <3>      3/9/00    gaw     Fix AEDesc.dataHandle references
         <2>      3/9/00    gaw     API changes for MoreAppleEvents
*/

//	Conditionals to setup the build environment the way we like it.
#include "MoreSetup.h"
//**************	Universal Headers		********************************
#include <AppleScript.h>
#include <ASDebugging.h>
#include <OSA.h>
#include <OSAGeneric.h>
#include <Resources.h>
//**************	ANSI Headers			********************************
#include <string.h>
//**************	Project Headers			********************************
#include "MoreAppleEvents.h"
#include "MoreOSA.h"
/********************************************************************************
	Checks for the presence of the AppleScript OSA component.
	
	RESULT CODES
	____________
	true		AppleScript is present and available
	false		It isn't
	____________
*/
pascal Boolean MoreOSAAppleScriptIsPresent(void)
{
	OSErr	anErr = noErr;
	
	static	long		gHasAppleScript = kFlagNotSet;
	
	if (gHasAppleScript == kFlagNotSet)
	{
		long	response;
		
		if (Gestalt(gestaltAppleScriptAttr, &response) == noErr)
		{
			gHasAppleScript = (response & (1L << gestaltAppleScriptPresent)) != 0;
		}
	}
	
	return gHasAppleScript;
}//end MoreAEHasAppleEvents
/********************************************************************************
	Open the default scripting component.

	componentInstance	input:	Pointer to ComponentInstance variable.
						output:	The newly opened scripting component, or nil if error.
	
	RESULT CODES
	____________
	noErr			 			0	No error	
	errOSACantOpenComponent	-1762	Can't connect to scripting system with that ID
	____________
*/
pascal OSErr MoreOSAOpenGenericScriptingComponent(ComponentInstance *compInstPtr)
{
	OSErr	anErr = noErr;
	
	if (MoreAEHasAppleEvents())
	{
		*compInstPtr = OpenDefaultComponent(kOSAComponentType, kOSAGenericScriptingComponentSubtype);
		if ((*compInstPtr == (ComponentInstance)badComponentInstance)
		  || (*compInstPtr == (ComponentInstance)badComponentSelector))
		{
			*compInstPtr = 0;					//	Set to 0 if no component is found.
			anErr = errOSACantOpenComponent;
		}
	}
	else
	{
		anErr = errOSACantOpenComponent;
	}
	
	return (anErr);
}//end MoreOSAOpenGenericScriptingComponent
/********************************************************************************
	Open the ApppleScript scripting component.

	componentInstance	input:	Pointer to ComponentInstance variable.
						output:	The newly opened scripting component, or nil if error.
	
	RESULT CODES
	____________
	noErr			 			0	No error	
	errOSACantOpenComponent	-1762	Can't connect to scripting system with that ID
	____________
*/
pascal OSErr MoreOSAOpenAppleScriptComponent(ComponentInstance *compInstPtr)
{
	OSErr	anErr = noErr;
	
	if (MoreAEHasAppleEvents())
	{
		*compInstPtr = OpenDefaultComponent(kOSAComponentType, kAppleScriptSubtype);
		if ((*compInstPtr == (ComponentInstance)badComponentInstance)
		  || (*compInstPtr == (ComponentInstance)badComponentSelector))
		{
			*compInstPtr = 0;					//	Set to 0 if no component is found.
			anErr = errOSACantOpenComponent;
		}
	}
	else
	{
		anErr = errOSACantOpenComponent;
	}
	
	return (anErr);
}//end MoreOSAOpenAppleScriptingComponent
/********************************************************************************
	Load a script resource for the given component instance.
	
	componentInstance	input:	Component instance to load script into.
	pRefNum			input:	Resource file to get script from.
	scriptResID			input:	Script (type 'scpt') resource ID to load.
	scriptIDPtr			input:	Pointer to OSAID to return the loaded script ID in.
						output:	The newly loaded script ID.
	
	RESULT CODES
	____________
	noErr			 				0	No error	
	resNotFound					 -192	Can't find the requested resource
	errOSACorruptData			-1702	Corrupt data	
	errOSASystemError			-1750	General scripting system error	
	errOSABadStorageType		-1752	Script data not for this scripting component
	errOSADataFormatObsolete	-1758	Data format is obsolete	
	errOSADataFormatTooNew		-1759	Data format is too new	
	badComponentInstance	$80008001	Invalid component instance	
	____________
*/
pascal OSErr MoreOSALoadScriptResource(ComponentInstance componentInstance, short pRefNum, short scriptResID, OSAID *scriptIDPtr)
{
	Handle		scriptHnd = nil;
	OSAError	anErr = noErr;
	short		originalResFile = CurResFile();
	
	UseResFile(pRefNum);
	
	*scriptIDPtr = kOSANullScript;		//	Set to null incase it can't be loaded
	
	scriptHnd = Get1Resource(typeScript, scriptResID);
	if (nil == scriptHnd)
	{
		anErr = resNotFound;
	}
	else
	{
		AEDesc	scriptDesc;
		
		scriptDesc.descriptorType = typeOSAGenericStorage;
		(void*) scriptDesc.dataHandle = (void*) scriptHnd;
		anErr = OSALoad(componentInstance, &scriptDesc, kOSAModeNull, scriptIDPtr);
		ReleaseResource(scriptHnd);	//	Once loaded, don't need the script resource any more.
	}
	
	UseResFile(originalResFile);
	
	return anErr;
}//end MoreOSALoadScriptResource
/********************************************************************************
	Dispose of a script for the given component instance.  If there have been
	changes to the script, it is saved before being unloaded.
	

	componentInstance	input:	Component instance to load script into.
	pRefNum			input:	Resource file to get script from.
	scriptResID			input:	Script (type 'scpt') resource ID to load.
	scriptIDPtr			input:	OSAID to return the loaded script ID in.
	
	RESULT CODES
	____________
	noErr			 				0	No error	
	memFullErr					 -108	Not enough memory	
	nilHandleErr				 -109	NIL master pointer	
	memWZErr					 -111	Attempt to operate on a free block	
	resNotFound					 -192	Can't find the requested resource
	resAttrErr					 198	Attribute inconsistent with operation
	errOSASystemError			-1750	General scripting system error	
	errOSAInvalidID				-1751	Invalid script ID	
	errOSABadStorageType		-1752	Desired type not supported by this scripting component   	
	errOSABadSelector			-1754	Selector value not supported by scripting component	
	badComponentInstance	$80008001	Invalid component instance	
	____________
*/
pascal OSErr MoreOSAUnloadScriptResource(ComponentInstance componentInstance, short pRefNum, short scriptResID, OSAID scriptID)
{
	OSAError	osaErr = noErr;
	
	long		scriptIsModified = 0;	
	short		originalResFile = CurResFile();
	
	UseResFile(pRefNum);
	osaErr = OSAGetScriptInfo(componentInstance, scriptID, kOSAScriptIsModified, &scriptIsModified);
	if (noErr == osaErr &&  scriptIsModified > 0)
	{
		Handle	scriptHnd = Get1Resource(typeScript, scriptResID);
		if (nil == scriptHnd)
		{
			osaErr = resNotFound;
		}
		else
		{
			AEDesc	scriptDesc = {typeNull, nil};
			
			osaErr = OSAStore(componentInstance, scriptID, typeOSAGenericStorage, kOSAModeNull, &scriptDesc);
			if (noErr == osaErr)
			{
				Size scriptSize = AEGetDescDataSize(&scriptDesc);

				SetHandleSize(scriptHnd, scriptSize);
				osaErr = MemError();
				if (noErr == osaErr)
				{
					osaErr = AEGetDescData(&scriptDesc,*scriptHnd,scriptSize);
					if (noErr == osaErr)
					{
						ChangedResource(scriptHnd);
						osaErr = ResError();
						if (noErr == osaErr)
						{
							UpdateResFile(pRefNum);
							osaErr = ResError();
						}
					}
				}
				(void) AEDisposeDesc(&scriptDesc);
			}
			ReleaseResource(scriptHnd);
		}
	}
	UseResFile(originalResFile);

	return osaErr;
}//end MoreOSAUnloadScriptResource

/********************************************************************************
	Given an AppleScript script ID, return an list containing the names of all
	handlers inplemented in the script.
	
	NOTE:  Valid only for AppleScript scripts.
	
	Esentially a wrapper around OSAGetHandlerNames() that allows you to pass in either
	a generic or AppleScript component instance.
	
	Returns a list of handler names (typeChar)
	
	The handler name for subroutines, sometimes called named handlers, for should be of 
	typeChar and be in cononical form, i.e., all lower case with diacriticals removed.  
	The name of a handler for a standard event (run, open, or event xxxxyyyy) will be 
	8 characters long, and will be of the form clasIDID where 'clas' is the event class 
	and 'IDID' is the eventID.

	
	compInst		input:	A generic or AppleScript component instance.
	scriptID		input:	The script to get the handler names from.
	nameDescPtr		input:	A list to contain the handler names.
	
	RESULT CODES
	____________
	____________
*/
pascal OSAError MoreOSAGetHandlerNames(ComponentInstance compInst, OSAID scriptID, AEDescList* handlerListPtr)
{
	OSAError			anErr = noErr;
	ComponentInstance	asCompInst;
	
	anErr = OSAGenericToRealID(compInst, &scriptID, &asCompInst);
	if (anErr == noErr)
	{
		anErr = OSAGetHandlerNames(asCompInst, kOSANullMode, scriptID, handlerListPtr);
	}
	return anErr;
}//end MoreOSAGetHandlerNames

/********************************************************************************
	Look for a particular handler name in a list of handler names (as returned by MoreOSAGetHandlerNames()).
	
	The handler name for subroutines, sometimes called named handlers, for should be of 
	typeChar and be in cononical form, i.e., all lower case with diacriticals removed.  
	The name of a handler for a standard event (run, open, or event xxxxyyyy) will be 
	8 characters long, and will be of the form clasIDID where 'clas' is the event class 
	and 'IDID' is the eventID.

	NOTE:  Valid only for AppleScript scripts.
	
	handlerListPtr		input:	The list of handler names for an AppleScript script.
	pRefNum			input:	The handler name to check for, in cononical form.
	nameDescPtr			input:	An string containing the name of the handler to look for
	
	RESULT CODES
	____________
	____________
*/
pascal Boolean MoreOSAHandlerIsInHandlerList(const AEDescList *handlerListPtr, const AEDesc *nameDescPtr)
{
	OSErr		anErr = noErr;
	Boolean		hasHandler = false;
	
	SInt32	itemCount;
		
	anErr = AECountItems(handlerListPtr, &itemCount);
	if (anErr == noErr)
	{
		UInt32		index;
		DescType	nameType = nameDescPtr->descriptorType;
		
		for (index = 1; index <= itemCount; index++)
		{
			SInt32		actualSize;
			char		handlerName[128];
			DescType	keyWord;
			DescType	actualType;		
			
			anErr = AEGetNthPtr(handlerListPtr, index, nameType, &keyWord, &actualType, 
								 &handlerName, sizeof(handlerName), &actualSize);
			if (anErr == noErr  &&  strncmp(handlerName, (char*)&(*nameDescPtr->dataHandle), actualSize) == 0)
			{
				hasHandler = true;
				break;
			}
		}
	}
	return hasHandler;
}//end MoreOSAHandlerIsInHandlerList

/********************************************************************************
	Check a script to see if it contains a particular handler.
	
	The handler name for subroutines, sometimes called named handlers, for should be of 
	typeChar and be in cononical form, i.e., all lower case with diacriticals removed.  
	The name of a handler for a standard event (run, open, or event xxxxyyyy) will be 
	8 characters long, and will be of the form clasIDID where 'clas' is the event class 
	and 'IDID' is the eventID

	For subroutines, the input should be of type 'text' with the dataHandle
	containing a lowercased name for AppleScript english scripts
	For other event handlers (open, quite, event clasIDID) the input should
	be of type 'evnt' with the data handle containing the event class and ID
	DescTypes, i.e., 8 bytes of the form [clasIDID]
	
	componentInstance	input:	Component instance for script.
	pRefNum			input:	Script to get names from.
	nameDescPtr			input:	A typeChar desc containing the name of the handler to look for
								Note: Must be in cononical form, i.e., 'MyHandler' would be 'myhandler'.
	
	Returns a boolean indicating if the handler was found.
	
	RESULT CODES
	____________
	____________
*/
pascal Boolean MoreOSAScriptHasHandler(ComponentInstance compInst, OSAID scriptID, const AEDesc *nameDescPtr)
{
	AEDescList	handlerList;
	OSErr		anErr = noErr;
	Boolean		hasHandler = false;
	
	anErr = MoreOSAGetHandlerNames(compInst, scriptID, &handlerList);
	if (anErr == noErr)
	{
		hasHandler = MoreOSAHandlerIsInHandlerList(&handlerList, nameDescPtr);
	}
	AEDisposeDesc(&handlerList);
	
	return hasHandler;
}//end MoreOSAScriptHasHandler
//********************************************************************************
